/* +------------------------------------------------------------------------+
   |                                                                        |
   |                    Centre de contrle mount-ppp                        |
   |                                                                        |
   +------------------------------------------------------------------------+ */

/* compilation :
   gcc -O2 -o xmount-ppp xmount-ppp.c -lX11 -lXpm -L/usr/X11R6/lib
*/
/* M. Quercia; le 29/04/2000 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

/* Pixmaps */
#include "cdrom.xpm"
#include "cdromro.xpm"
#include "disquette.xpm"
#include "disquettero.xpm"
#include "disquetterw.xpm"
#include "zip.xpm"
#include "zipro.xpm"
#include "ziprw.xpm"
#include "PhoneOld.xpm"
/* toutes les pixmaps doivent avoir les mmes dimensions */
#define pmwidth  48
#define pmheight 48

                   /* +-------------------------+
                      |  Structures de donnes  |
                      +-------------------------+ */

/* pixmap */
typedef struct {
  Pixmap pixels,mask;
  XpmAttributes attr;
} pxmap; 

/* objet abstrait */
typedef struct {
  void *next;                            /* objet suivant        */
  Window w;                              /* fentre associe     */
  void (*event)(void *data, XEvent *e);  /* gestion d'vnements */
  void (*check)(void *data);             /* contrle priodique  */
} object;

/* fentre d'information */
typedef struct {
  object ob;
  char *texte;            /* texte de l'info-bulle */
} info;

/* disque */
typedef enum {Unmounted, MountedRO, MountedRW} disk_state;
typedef struct {
  object ob;
  disk_state etat;         /* tat de montage   */
  pxmap *px, *pxro, *pxrw; /* pixmaps associes */
  char  *mpoint;           /* point de montage  */
  char  *eject;            /* nom pour eject    */
  int   device;            /* n majeur-mineur  */
  int   press;             /* bouton enfon    */
} disque;

/* ligne ppp */
typedef enum {Unconnected, Connecting, Connected} ppp_state;
typedef struct {
  void *next;             /* ligne suivante       */
  char *nom;              /* nom de la connexion  */
  int  adr;               /* adresse passerelle   */
  int  mask;              /* masque rseau        */
  ppp_state etat;         /* tat de la connexion */
} ppp;

/* tlphone */
typedef struct {
  object ob;
  pxmap *px;              /* pixmap de fond       */
  ppp   *ppp;             /* liste des lignes     */
  ppp   *act;             /* ligne active         */
  int   show;             /* menu visible         */
  int   press;            /* bouton enfon       */
  void  *menu;            /* menu  afficher      */
} phone;

/* menu */
typedef struct {
  object ob;
  phone  *ph;             /* tlphone associ   */
  int    cur;             /* n ligne courante   */
  int    max;             /* nb de lignes        */
  int    w,h;             /* dimension fentre   */
} menu;

/* liste d'options */
typedef struct {
  void *next;
  char *key;
  char *val;
} option;

/* donnes relatives au serveur X */
Display       *display;
int           screen;
Window        mainw;
XSizeHints    indic;
XSetWindowAttributes infoattr;
GC            gc;
Colormap      cmap;
unsigned long bg,fg,rg,yg,sg,hg;
XFontStruct   *font;
Cursor        cursor;
int           dy,y0;
object        *objects;
info          *inf;
option        *options;

/* commandes  lancer */
char *mountcmd, *umountcmd, *ejectcmd, *callcmd;

/* donnes spcifiques au tlphone */
char *phdev, *getmail, *pppd;


                          /* +-----------+
                             |  Options  |
                             +-----------+ */


/* lecture du fichier de configuration */
option *parse_config(option *p) {

  option *q = p;
  char nom[256];
  char ligne[256];
  char *key, *val, *end;
  char *path = getenv("HOME");
  FILE *f;
  int l,num;

  if (path) {
    sprintf(nom,"%s/.xmount-ppprc",path);
#ifdef debug
    fprintf(stderr,"config = %s\n", nom);
#endif
    f = fopen(nom,"r");
    if (f) {

      num = 0;
      while (fgets(ligne,256,f) != NULL) {
	num++;
	key = ligne;
	l = strlen(key);
	if ((l > 0) && (key[l-1] == '\n')) key[l-1] = '\0';
#ifdef debug
	fprintf(stderr,"%d: %s\n",num, ligne);
#endif

	/* saute les espaces en dbut de ligne */
	while (isblank(*key)) key++;
	if ((*key != '\0') && (*key != '#')) {

	  /* cherche la fin de la cl */
	  val = key;
	  while ((*val) && !isblank(*val)) val++;
	  if (*val == '\0') {
	    fprintf(stderr,"%s: ligne %d incorrecte\n", nom, num);
	    exit(1);
	  }
	  *val = 0;
	  val++;

	  /* dlimite la valeur */
	  while (isblank(*val)) val++;
	  end = val;
	  while ((*end) && !isblank(*end)) end++;
	  *end = 0;

	/* enregistre le couple */
	  q = (option *)malloc(sizeof(option));
	  if (q == NULL) {
	    fprintf(stderr,"mmoire insuffisante\n");
	    exit(1);
	  }
	  q->next = p;
	  q->key  = (char *)malloc(strlen(key)+1);
	  if (q->key == NULL) {
	    fprintf(stderr,"mmoire insuffisante\n");
	    exit(1);
	  }
	  strcpy(q->key,key);
	  q->val  = (char *)malloc(strlen(val)+1);
	  if (q->val == NULL) {
	    fprintf(stderr,"mmoire insuffisante\n");
	    exit(1);
	  }
	  strcpy(q->val,val);
	  p = q;
	}
      }
      fclose(f);
    }
  }
#ifdef debug
  fflush(stderr);
#endif
  return(q);
}
	
/* analyse la ligne de commande */
option *parse_argv(option *p, int argc, char **argv) {
  option *q;

  for (; argc > 1; argc-=2, argv+=2) {
    q = (option *)malloc(sizeof(option));
    if (q == NULL) {
      fprintf(stderr,"mmoire insuffisante\n");
      exit(1);
    }
    q->next = p;
    q->key  = (argv[0][0] == '-') ? argv[0]+1 : argv[0];
    q->val  = argv[1];
    p = q;
  }

  if (argc) {
    fprintf(stderr,"nombre impair d'arguments\n");
    exit(1);
  }

  return(p);
}
    
/* extrait une option */
char *get_option(char *key, char *defaut) {
  option **p;
  char *res = defaut;

  for (p = &options; (*p != NULL) && (strcmp((*p)->key,key) != 0);
       p = (option **)&((*p)->next));
  if (*p) {
    res = (*p)->val;
    *p = (*p)->next;
  }
#ifdef debug
  fprintf(stderr,"option(%s) = %s\n",key,res);
  fflush(stderr);
#endif
  return(res);
}

/* description d'un priphrique */
void parse_device(char *desc, char **mpoint, int *device, char **eject) {
  struct stat sts;
  char *dev;

  *mpoint = strchr(desc,',');
  if (*mpoint == NULL) {
    fprintf(stderr,"priphrique incorrect: %s\n",desc);
    exit(1);
  }
  **mpoint = 0;
  (*mpoint)++;

  dev = strchr(*mpoint,',');
  if (dev == NULL) {
    fprintf(stderr,"priphrique incorrect: %s\n",*mpoint);
    exit(1);
  }
  *dev = 0;
  dev++;

  *eject = strchr(dev,',');
  if (*eject != NULL) {**eject = 0; (*eject)++;}

  if (stat(dev, &sts) != 0) {
    fprintf(stderr, "%s: accs impossible\n", dev);
    exit(1);
  }
  if (sts.st_rdev == 0) {
    fprintf(stderr, "%s n'est pas un device valide\n", dev);
    exit(1);
  }
  *device = sts.st_rdev;
}

/* description d'une ligne ppp */
int mtab[] = {
  0x0,       0x80,       0xc0,       0xe0,       0xf0,       0xf8,       0xfc,       0xfe,       0xff,
  0x0ff,     0x80ff,     0xc0ff,     0xe0ff,     0xf0ff,     0xf8ff,     0xfcff,     0xfeff,     0xffff,
  0x0ffff,   0x80ffff,   0xc0ffff,   0xe0ffff,   0xf0ffff,   0xf8ffff,   0xfcffff,   0xfeffff,   0xffffff,
  0x0ffffff, 0x80ffffff, 0xc0ffffff, 0xe0ffffff, 0xf0ffffff, 0xf8ffffff, 0xfcffffff, 0xfeffffff, 0xffffffff
};
void parse_ppp(char *desc, int *adr, int *mask) {
  char *sadr;
  int a,b,c,d,m;

  /* cherche le dbut de l'adresse */
  sadr = strchr(desc,',');
  if (sadr == NULL) {
    fprintf(stderr,"lignee incorrecte: %s\n",desc);
    exit(1);
  }
  *sadr = 0;
  sadr++;

  if ((sscanf(sadr,"%d.%d.%d.%d/%d",&a,&b,&c,&d,&m) != 5) || (m < 0) || (m > 32))
    {
      fprintf(stderr,"adresse incorrecte: %s\n",sadr);
      exit(1);
    }

  *adr = a + (b << 8) + (c << 16) + (d << 24);
  *mask = mtab[m];
}

/* gomtrie */
void parse_geometry(char *desc, int *x, int *y, int *sx, int *sy) {
  char *s = desc;

  /* signe de x */
  while ((*s != '\0') && (*s != '+') && (*s != '-')) s++;
  switch(*s) {
  case '+': *sx = 1;  s++; break;
  case '-': *sx = -1; s++; break;
  default: fprintf(stderr,"gomtrie incorrecte: %s\n",desc); exit(1);
  }

  /* valeur de x */
  for (*x = 0; isdigit(*s); s++) *x = 10*(*x) + (s[0]-'0');

  /* signe de y */
  switch(*s) {
  case '+': *sy = 1;  s++; break;
  case '-': *sy = -1; s++; break;
  default: fprintf(stderr,"gomtrie incorrecte: %s\n",desc); exit(1);
  }

  /* valeur de y */
  for (*y = 0; isdigit(*s); s++) *y = 10*(*y) + (s[0]-'0');

  if (*s) {
    fprintf(stderr,"gomtrie incorrecte: %s\n",desc);
    exit(1);
  }

}

                   /* +------------------------+
                      |  Communication avec X  |
                      +------------------------+ */

/* allocation de couleur */
unsigned long color(char *name) {
  XColor couleur,exacte; 
  XAllocNamedColor(display,cmap,name,&couleur,&exacte);
  return(couleur.pixel);
}

/* affiche un rectangle en relief */
void draw_rect(Window win, int creux, int x, int y, int w, int h) {

  XPoint ligne[3];

  XSetClipMask(display,gc,None);
  XSetForeground(display,gc,(creux == 0) ? hg : (creux == 1) ? sg : bg);
  ligne[0].x = x;
  ligne[0].y = y+h-1;
  ligne[1].x = 0;
  ligne[1].y = 1-h;
  ligne[2].x = w-1;
  ligne[2].y = 0;
  XDrawLines(display,win,gc,ligne,3,CoordModePrevious);

  XSetForeground(display,gc,(creux == 0) ? sg : (creux == 1) ? hg : bg);
  ligne[1].x = w-1;
  ligne[1].y = 0;
  ligne[2].x = 0;
  ligne[2].y = 1-h;
  XDrawLines(display,win,gc,ligne,3,CoordModePrevious);

}

/* dessin de la fentre principale */
void main_event(void *data, XEvent *e) {
  switch(e->type) {
  case Expose:
    draw_rect(mainw,0,1,1,indic.width-2,indic.height-2);
    draw_rect(mainw,1,2,2,indic.width-4,indic.height-4);
    break;
  default: break;
  }
}

/* affichage d'une chane en bandeau */
void draw_string(Window w, int y, char *s, unsigned long couleur) {

  int l = strlen(s);
  int dir,asc,desc;
  XCharStruct over;

  XTextExtents(font,s,l,&dir,&asc,&desc,&over);
  XSetClipMask(display,gc,None);
  XSetForeground(display,gc,yg);
  XFillRectangle(display,w,gc,0,y-asc,pmwidth+4,asc+desc); 
  XSetForeground(display,gc,couleur);
  XDrawString(display,w,gc,0,y,s,l);
}

/* connection au serveur X */
object *connecte(int argc, char **argv, object *p,
		 int x, int y, int sx, int sy, int w, int h) {

  int rx,ry,rw,rh,bd,depth;
  Window r;
  char *nom;
  object *q;

  /* isole le nom du programme */
  nom = strrchr(argv[0],'/');
  nom = (nom == NULL) ? argv[0] : nom+1;

  /* connection au serveur X */
  display = XOpenDisplay(NULL);
  if (display == NULL)
    {
      fprintf(stderr, "display inaccessible\n");
      exit(1);
    }
  screen = DefaultScreen(display);
  cursor = XCreateFontCursor(display,XC_hand2);

  /* initialisation des couleurs */
  cmap = DefaultColormap(display,screen);
  bg = color(get_option("bg","#ddddd0"));
  fg = color(get_option("fg","black"));
  rg = color(get_option("rg","red"));
  yg = color(get_option("yg","yellow"));
  sg = color(get_option("sg","black"));
  hg = color(get_option("hg","white"));

  /* fonte */
  gc = XCreateGC(display,DefaultRootWindow(display),0,NULL);
  XSetBackground(display,gc,bg);
  font = XLoadQueryFont(display,get_option("fn","-adobe-helvetica-medium-r-normal--14-140-75-75-p-77-iso8859-1"));
  XSetFont(display,gc,font->fid);
  dy = font->ascent + font->descent + 2;
  y0 = font->ascent;

  /* calcule la position par rapport  la fentre racine */
  XGetGeometry(display,DefaultRootWindow(display),&r,&rx,&ry,&rw,&rh,&bd,&depth);
  indic.width = w;
  indic.height = h;
  indic.x = (sx > 0) ? x : rw-w-x;
  indic.y = (sy > 0) ? y : rh-h-y;
  indic.flags = PPosition;
#ifdef debug
  fprintf(stderr, "g = %dx%d+%d+%d\n",indic.width,indic.height,indic.x,indic.y);
  fflush(stderr);
#endif

  /* ouvre la fentre principale */
  mainw = XCreateSimpleWindow(
	     display,
	     DefaultRootWindow(display),
	     indic.x, indic.y, indic.width, indic.height,
	     0, fg, bg);
  XSetStandardProperties(display, mainw, nom, nom, None, argv, argc, &indic);
  XDefineCursor(display,mainw,cursor);
  XMapRaised(display,mainw);
  XSelectInput(display, mainw, ExposureMask);

  /* insertion dans la liste d'objets */
  q = (object *)malloc(sizeof(object));
  if (q == NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }
  q->next = p;
  q->w = mainw;
  q->event = &main_event;
  q->check = NULL;
  return(q);

}

/* allocation d'une pixmap */
void make_pixmap(char **table, pxmap *px) {
  memset(&(px->attr), 0, sizeof(XpmAttributes));
  XpmCreatePixmapFromData(
     display, mainw,
     table,  &(px->pixels), &(px->mask), &(px->attr)
     );
}

/* rception d'un vnement */
void xevent(object *p) {
  XEvent e;
  XNextEvent(display, &e);
  for (; (p != NULL) && (p->w != e.xany.window); p = p->next);
  if ((p != NULL) && (p->event != NULL)) p->event(p,&e);
}

                             /* +--------------+
                                |  Info-bulle  |
                                +--------------+ */


/* affiche le texte */
void info_event(void *data, XEvent *e) {
  switch(e->type) {
  case Expose:
    XSetClipMask(display,gc,None);
    XSetForeground(display,gc,yg);
    XFillRectangle(display,inf->ob.w,gc,0,0,1000,1000);
    XSetForeground(display,gc,fg);
    if (inf->texte)
      XDrawString(display,inf->ob.w,gc,1,y0+1,inf->texte,strlen(inf->texte));
    break;
  default: break;
  }
}

/* activation */
void info_show(Window win, char *msg) {

  int dir,asc,desc,x,y,w,h;
  XCharStruct over;
  Window child;

  inf->texte = msg;
  XTextExtents(font,msg,strlen(msg),&dir,&asc,&desc,&over);
  w = over.width+2;
  h = asc+desc+2;

  XTranslateCoordinates(display,win,DefaultRootWindow(display),
			(pmwidth+4-w)/2,-h-6,&x,&y,&child);
  if (y < 0)
  XTranslateCoordinates(display,win,DefaultRootWindow(display),
			(pmwidth+4-w)/2,pmheight+4+4,&x,&y,&child);

  XMoveResizeWindow(display,inf->ob.w,x,y,w,h);
  XMapRaised(display,inf->ob.w);
}

/* dsactivation */
void info_remove() {
  XUnmapWindow(display,inf->ob.w);
}

/* cration */
object *info_add(object *p) {

  inf = (info *)malloc(sizeof(info));
  if (inf == NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }

  infoattr.override_redirect = 1;
  inf->ob.w = XCreateWindow(
	     display,
	     DefaultRootWindow(display),
	     0,0,10,10,
             1,
	     DefaultDepth(display,screen),
	     CopyFromParent,
	     CopyFromParent,
	     CWOverrideRedirect,&infoattr);
  XSelectInput(display, inf->ob.w, ExposureMask);
  inf->ob.next = p;
  inf->ob.event = &info_event;
  inf->ob.check = NULL;
  inf->texte = NULL;
  return((object *)inf);
}

                               /* +----------+
                                  |  Disque  |
                                  +----------+ */

/* affiche l'icne approprie */
void disk_draw(disque *d) {
  pxmap *px;

  /* dessine le contour */
  draw_rect(d->ob.w, d->press, 0,0,pmwidth+4,pmheight+4);

  /* dessine la pixmap */
  switch(d->etat) {
  case Unmounted: px = d->px;   break;
  case MountedRO: px = d->pxro; break;
  case MountedRW: px = d->pxrw; break;
  }

  XSetClipMask(display,gc,px->mask);
  XSetClipOrigin(display,gc,2,2);
  XCopyArea(display,
	    px->pixels,
	    d->ob.w,
	    gc,
	    0,0, px->attr.width,px->attr.height, 2,2
	    );

}

/* monte ou dmonte le disque */
void mount(disque *d, char *cmd) {
  if (fork() == 0) {execl(cmd,cmd,d->mpoint,NULL); exit(1);}
  disk_draw(d);
}

/* jecte un disque */
void eject(disque *d) {
  if (fork() == 0) {
    if (d->etat != Unmounted) {
      if (fork() == 0) {execl(umountcmd,umountcmd,d->mpoint,NULL); exit(1);}
      else wait(NULL);
    }
    if (d->eject) execl(ejectcmd,ejectcmd,d->eject,NULL);
    else exit(0);
  }
  disk_draw(d);
}

/* rception d'un vnement X */
void disk_event(disque *d, XEvent *e) {
  switch(e->type) {
  case ButtonPress:
    d->press=1;
    disk_draw(d);
    break;
  case ButtonRelease: 
    if (d->press) {
      d->press = 0;
      if (e->xbutton.button == 3) eject(d);
      else mount(d, (d->etat == Unmounted) ? mountcmd : umountcmd);
    }
    break;
  case Expose:
    if (e->xexpose.count == 0) disk_draw(d);
    break;
  case EnterNotify:
    info_show(d->ob.w,d->mpoint);
    d->press = ((e->xcrossing.state & 0x700) != 0);
    disk_draw(d);
    break;
  case LeaveNotify:
    info_remove();
    d->press=0;
    disk_draw(d);
    break;
  default: break;
  }
}

/* contrle le point de montage */
void disk_check(disque *d) {
  static struct stat sts;
  int i;
  disk_state new_etat;

  if (stat(d->mpoint, &sts) != 0) {
      fprintf (stderr, "%s: accs impossible\n", d->mpoint);
      exit (1);
    }
    new_etat = (sts.st_dev != d->device) ? Unmounted :
               (access(d->mpoint, W_OK)) ? MountedRO : MountedRW;
    if (new_etat != d->etat) {
      d->etat = new_etat;
      disk_draw(d);
      XFlush(display);
    }
}

/* assemblage d'une structure disque */
object *disk_add(p, x, px,pxro,pxrw, mpoint,device,eject)
     object *p;                  /* liste en cours     */
     int    x;                   /* numro fentre     */
     pxmap  *px, *pxro, *pxrw;   /* pixmaps  afficher */
     char   *mpoint;             /* point de montage   */
     int    device;              /* priph.  monter   */
     char   *eject;              /* param. pour eject  */
{
  disque *d = (disque *)malloc(sizeof(disque));
  if (d==NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }
  d->ob.next = p;
  d->ob.w = XCreateSimpleWindow(
		display,mainw,
		(6+pmwidth)*x+4, 4, pmwidth+4, pmheight+4,
		0, fg, bg);
  XSelectInput(display, d->ob.w,
	       ButtonPressMask
	       |ButtonReleaseMask
	       |ExposureMask
	       |EnterWindowMask
	       |LeaveWindowMask);
  XMapWindow(display,d->ob.w);
  (void *)d->ob.event = &disk_event;
  (void *)d->ob.check = &disk_check;
  d->etat   = Unmounted;
  d->px     = px;
  d->pxro   = pxro;
  d->pxrw   = pxrw;
  d->mpoint = mpoint;
  d->eject  = eject;
  d->device = device;
  d->press  = 0;
  return((object *)d);
}

                              /* +------------+
                                 |  Menu ppp  |
                                 +------------+ */


/* affichage */
void menu_draw(menu *m) {
  ppp *p;
  int y;

  /* efface la fentre */
  XSetClipMask(display,gc,None);
  XSetForeground(display,gc,bg);
  XFillRectangle(display,m->ob.w,gc,0,0,1000,1000);

  /* affiche les lignes */
  XSetForeground(display,gc,fg);
  for (p=m->ph->ppp, y=y0+2; p != NULL; p=(ppp *)p->next, y+=dy) {
    XDrawImageString(display,m->ob.w,gc,4,y,p->nom,strlen(p->nom));
  }

  /* contours */
  draw_rect(m->ob.w,0,1,1,m->w-2,m->h-2);
  if (m->cur >= 0) {
    draw_rect(m->ob.w,0,2,(m->cur)*dy+2,m->w-4,dy);
  } 
}


/* dplace le bandeau avec la souris */
void menu_follow(menu *m, int x, int y) {
  int c;
  if ((x < 0) || (x >= m->w) || (y < 0) || (y >= m->h)) c = -1;
  else c = (y-2)/dy;
  if ((c < 0) || (c > m->max)) c = -1;
  if (c != m->cur) {
    if (m->cur >= 0) draw_rect(m->ob.w,2,2,(m->cur)*dy+2,m->w-4,dy);
    m->cur = c;
    if (m->cur >= 0) draw_rect(m->ob.w,0,2,(m->cur)*dy+2,m->w-4,dy);
  }
}

/* activation */
void menu_show(menu *m) {
  int x,y;
  Window child;

  XTranslateCoordinates(display,m->ph->ob.w,DefaultRootWindow(display),
			0,-m->h-6,&x,&y,&child);
  if (y < 0)
    XTranslateCoordinates(display,m->ph->ob.w,DefaultRootWindow(display),
			0,pmheight+4+4,&x,&y,&child);

  m->cur = -1;
  XMoveWindow(display,m->ob.w,x,y);
  XMapRaised(display,m->ob.w);
  XGrabPointer(display, m->ob.w, 1,
	       PointerMotionMask|ButtonPressMask,
	       GrabModeAsync,
	       GrabModeAsync,
	       None,
	       None,
	       CurrentTime);
}

/* dsactivation */
void menu_remove(menu *m) {
  XUnmapWindow(display,m->ob.w);
}

/* active une connexion */
void phone_draw(int clear, phone *p);
void menu_call(menu *m, int button) {
  ppp *q;
  int c;

  if ((m->cur >= 0) && (m->cur < m->max) && (m->ph->act == NULL)) {
    for(c=m->cur, q=m->ph->ppp; c>0; q = (ppp *)q->next, c--);
    m->ph->act = q;
    q->etat = Connecting;
    phone_draw(1,m->ph);

    if (fork() == 0) {
      int fd = open("/dev/null",O_WRONLY);

      /* bouton 3 -> getmail */
      if (button == 3) {
	FILE *pidfile = fopen(getmail,"w");
	if (pidfile) {
	  fprintf(pidfile,"%d",getpid());
	  fclose(pidfile);
	}
      }

      dup2(fd,1);
      dup2(fd,2);
      execl(callcmd, callcmd, q->nom, NULL);
      exit(1);
    }
  }
}

/* rception d'un vnement X */
void menu_event(menu *m, XEvent *e) {
  switch(e->type) {
  case Expose:
    if (e->xexpose.count == 0) menu_draw(m);
    break;
  case MotionNotify:
    menu_follow(m,e->xmotion.x,e->xmotion.y);
    break;
  case ButtonPress:
    menu_call(m,e->xbutton.button);
    m->ph->show = 0;
    menu_remove(m);
    break;
  default: break;
  }
}

/* cration */
object *menu_add(object *p, phone *ph) {

  ppp *q;
  int dir,asc,desc;
  XCharStruct over;

  menu *m = (menu *)malloc(sizeof(menu));
  if (m == NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }

  /* nombre de lignes et taille du menu */
  for (q=ph->ppp, m->max=0, m->w = pmwidth-4; q != NULL; q = (ppp *)q->next) {
    m->max++;
    XTextExtents(font,q->nom,strlen(q->nom),&dir,&asc,&desc,&over);
    if (m->w < over.width) m->w = over.width;
  }
  m->w += 8;
  m->h = (m->max)*dy+4;

  /* cre la fentre */
  infoattr.override_redirect = 1;
  m->ob.w = XCreateWindow(
	     display,
	     DefaultRootWindow(display),
	     0,0,m->w,m->h,
             1,
	     DefaultDepth(display,screen),
	     CopyFromParent,
	     CopyFromParent,
	     CWOverrideRedirect,&infoattr);
  XDefineCursor(display,m->ob.w,cursor);
  XSelectInput(display, m->ob.w, ExposureMask|PointerMotionMask);

  /* chanage */
  m->ob.next = p;
  (void *)m->ob.event = (void *)&menu_event;
  m->ob.check = NULL;
  m->ph = ph;
  m->cur = -1;
  ph->menu = m;
  return((object *)m);
}

                              /* +-------------+
                                 |  Tlphone  |
                                 +-------------+ */


/* affichage */
void phone_draw(int clear, phone *p) {
  ppp *q;
  int y;

  /* efface la fentre si demand */
  if (clear) {
    XSetClipMask(display,gc,None);
    XSetForeground(display,gc,bg);
    XFillRectangle(display,p->ob.w,gc,0,0,1000,1000); 
  }

  /* dessine le contour */
  draw_rect(p->ob.w, p->press, 0,0,pmwidth+4,pmheight+4);

  /* affiche l'icne et la ligne active */
  if ((!p->show) || (p->act)) {
    XSetClipMask(display,gc,p->px->mask);
    XSetClipOrigin(display,gc,0,0);
    XCopyArea(display,
	      p->px->pixels,
	      p->ob.w,
	      gc,
	      0,0, p->px->attr.width,p->px->attr.height, 2,2
	      );
    if (p->act) {
      draw_string(p->ob.w,pmheight,p->act->nom,
		  (p->act->etat == Connected) ? rg : fg);
    }
  }
}

/* coupe la connexion active */
void phone_hangup(phone *p) {

  FILE *pppdfile = fopen(pppd,"r");
  if (pppdfile) {
    int pid;
    fscanf(pppdfile,"%d",&pid);
    kill(pid,SIGINT);
    fclose(pppdfile);
  }

  p->act->etat = Unconnected;
  p->act = NULL;
  phone_draw(1,p);
}

/* rception d'un vnement X */
void phone_event(phone *p, XEvent *e) {
  switch(e->type) {
  case ButtonPress:
    p->press = 1;
    phone_draw(0,p);
    break;
  case ButtonRelease:
    if (p->press) {
      p->press = 0;
      if (p->act) phone_hangup(p);
      else {
	phone_draw(0,p);
	if (p->show) {p->show = 0; menu_remove(p->menu);}
	else {p->show = 1;  menu_show(p->menu);}
      }
    }
    break;
  case Expose:
    if (e->xexpose.count == 0) phone_draw(0,p);
    break;
  case EnterNotify:
    info_show(p->ob.w,"ppp");
    p->press = ((e->xcrossing.state & 0x700) != 0);
    phone_draw(0,p);
    break;
  case LeaveNotify:
    info_remove();
    p->press = 0;
    phone_draw(0,p);
    break;
  default: break;
  }
}

/* teste si une ligne est active */
void phone_check(phone *p) {
  char buff[200];
  long adr;
  ppp  *q = NULL;
  FILE *f;

  /* examine /proc/net/route pour voir si l'interface est active */
  f = fopen("/proc/net/route", "r");
  while (fgets(buff,sizeof(buff),f) != NULL) {
    if (strncmp(phdev,buff,strlen(phdev)) == 0) {
      
      /* si active, compare l'adresse distante aux adresses enregistres */
      sscanf(buff+strlen(phdev), "%x", &adr);
      for (q = p->ppp;
	   (q != NULL) && ((q->mask & (q->adr ^ adr)) != 0);
	   q = (ppp *)q->next);
      break;
    }
  }
  fclose(f);

  /* met  jour l'affichage */
  if (q == NULL) {
    if ((p->act) && (p->act->etat != Connecting)) {
      p->act->etat = Unconnected;
      p->act = NULL;
      phone_draw(1,p);
      XFlush(display);
    }
  }

  else {
    if ((p->act == NULL) || (p->act->etat != Connected)) {
      p->act = q;
      q->etat = Connected;
      phone_draw(1,p);
      XFlush(display);
    }
  }
}

/* assemblage d'une structure phone */
object *phone_add(object *p, int x, pxmap *px) {

  phone *d = (phone *)malloc(sizeof(phone));
  if (d==NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }
  d->ob.next = p;
  d->ob.w = XCreateSimpleWindow(
		display,mainw,
		(6+pmwidth)*x+4, 4, pmwidth+4, pmheight+4,
		0, fg, bg);
  XSelectInput(display, d->ob.w,
	       ButtonPressMask
	       |ButtonReleaseMask
	       |ExposureMask
	       |EnterWindowMask
	       |LeaveWindowMask);
  XMapWindow(display,d->ob.w);
  (void *)d->ob.event = &phone_event;
  (void *)d->ob.check = &phone_check;
  d->px   = px;
  d->ppp  = NULL;
  d->act  = NULL;
  d->show = 0;
  return((object *)d);
}

/* adjonction d'une ligne ppp */
void ppp_add(p, nom,adr,mask)
     phone *p;               /* descr. tlphone     */
     char *nom; 	     /* nom de la connection */
     int adr;   	     /* adresse passerelle   */
     int mask; 	             /* masque rseau        */
{
  ppp *q = (ppp *)malloc(sizeof(ppp));
  if (q==NULL) {
    fprintf(stderr,"mmoire insuffisante\n");
    exit(1);
  }
  q->next = p->ppp;
  q->nom  = nom;
  q->adr  = adr;
  q->mask = mask;
  q->etat = Unconnected;
  p->ppp  = q;
}


                    /* +-----------------------+
                       |  Gestion des signaux  |
                       +-----------------------+ */

/* alarme priodique */
void handle_alarm(int sig) {
  object *p;
  for (p=objects; p!=NULL; p=p->next) if (p->check != NULL) p->check(p);
  signal(SIGALRM,handle_alarm);
  alarm(1);
}

/* prend note de la mort d'un processus fils */
void child_dead(int sig) {
  wait(NULL);
  signal(SIGCHLD,child_dead);
}


                    /* +-----------------------+
                       |  Programme principal  |
                       +-----------------------+ */


main(int argc, char **argv)
{
  pxmap  diskcd,      diskcdro,
         diskfloppy,  diskfloppyro, diskfloppyrw,
         diskzip,     diskzipro,    diskziprw,
         phoneold;
  int i,device,adr,mask;
  int x,y,sx,sy;
  char *obj,*mpoint,*eject;
  option *p;

  /* traitement des options */
  options = parse_config(NULL);
  options = parse_argv(options,argc-1,argv+1);
#ifdef debug
  for (p = options; p != NULL; p = (option *)p->next)
    fprintf(stderr,"option: %s -> %s\n",p->key,p->val);
  fflush(stderr);
#endif
  mountcmd  = get_option("mount", "/bin/mount");
  umountcmd = get_option("umount","/bin/umount");
  ejectcmd  = get_option("eject","/usr/bin/eject");
  callcmd   = get_option("dial",  "/usr/bin/wvdial");
  phdev     = get_option("ppp",   "ppp0");
  getmail   = get_option("getmailpid", "/var/run/getmail.pid");
  pppd      = get_option("pppdpid",    "/var/run/ppp0.pid");


  /* compte le nombre de pixmaps et dtermine la gomtrie */
  i=0;
  for (p = options; p != NULL; p = (option *)p->next)
    if (strcmp(p->key,"device") == 0) i++;
  for (p = options; p != NULL; p = (option *)p->next)
    if (strcmp(p->key,"line") == 0) {i++; break;}
  parse_geometry(get_option("geometry","-0+0"),&x,&y,&sx,&sy);

  /* connection au serveur X et initialisation des pixmaps */
  objects = connecte(argc,argv,NULL,x,y,sx,sy,(6+pmwidth)*i+6, pmheight+12);
  make_pixmap( cdrom_xpm,       &diskcd);
  make_pixmap( cdromro_xpm,     &diskcdro);
  make_pixmap( disquette_xpm,   &diskfloppy);
  make_pixmap( disquettero_xpm, &diskfloppyro);
  make_pixmap( disquetterw_xpm, &diskfloppyrw);
  make_pixmap( zip_xpm,         &diskzip);
  make_pixmap( zipro_xpm,       &diskzipro);
  make_pixmap( ziprw_xpm,       &diskziprw);
  make_pixmap( PhoneOld,        &phoneold);
  objects = info_add(objects);
  i = 0;

  /* priphriques  surveiller */
  obj=get_option("device","");
  while (obj[0]) {
    parse_device(obj,&mpoint,&device,&eject);
    if (strcmp(obj,"disquette") == 0)
      objects = disk_add(objects,i, &diskfloppy, &diskfloppyro, &diskfloppyrw, mpoint, device, eject);
    else if (strcmp(obj,"cdrom") == 0)
      objects = disk_add(objects,i, &diskcd,     &diskcdro,     &diskcdro,     mpoint, device, eject);
    else if (strcmp(obj,"zip") == 0)
      objects = disk_add(objects,i, &diskzip,    &diskzipro,    &diskziprw,    mpoint, device, eject);
    else {
      fprintf(stderr,"icne inconnue: %s\n",obj);
      exit(1);
    }
    obj=get_option("device","");
    i++;}
  
  /* lignes ppp */
  obj = get_option("line","");
  if (obj[0]) {
    objects = phone_add(objects, i, &phoneold);
    i++;
    while (obj[0]) {
      parse_ppp(obj, &adr, &mask);
      ppp_add(objects, obj, adr, mask);
      obj = get_option("line","");
    }
    objects = menu_add(objects, (phone *)objects);
  }

  /* vrifie qu'il y a qqch  faire */
  if (i == 0) {
    fprintf(stderr,"aucune action spcifie\n");
    exit(1);
  }

  /* dmarre le contrle priodique */
  handle_alarm(0);
  signal(SIGCHLD,child_dead);

  /* boucle de gestion des vnements */
  while(1) xevent(objects);
}

/*---------------------------- Termin ----------------------------------*/
